import io
import unittest
from datetime import datetime, timedelta
from unittest.mock import Mock

from minio import Minio

from common.utils.baseline.core import BaselineError, BaselineService
from common.utils.baseline.data import Data, PartialBaseline
from common.utils.common_model import Interval


class TestCore(unittest.TestCase):
    def setUp(self):
        self.baseline_service = BaselineService()

    def test_core_error_wrapping(self):
        self.assertRaises(BaselineError, self.baseline_service.read_all_partial_baselines, "vscan", "1234", "2541")

    def test_get_missing_baseline_intervals(self):
        start_time = datetime(year=2023, month=1, day=1)
        expected_interval = Interval(start_time, start_time + timedelta(days=29))

        current_intervals_1 = [
            Interval(start_time + timedelta(days=2), start_time + timedelta(days=3)),
            Interval(start_time + timedelta(days=3), start_time + timedelta(days=4)),
            Interval(start_time + timedelta(days=4), start_time + timedelta(days=20)),
            Interval(start_time + timedelta(days=23), start_time + timedelta(days=25)),
        ]

        missing_intervals_1 = [
            Interval(start_time, start_time + timedelta(days=2)),
            Interval(start_time + timedelta(days=20), start_time + timedelta(days=23)),
            Interval(start_time + timedelta(days=25), start_time + timedelta(days=29)),
        ]

        current_intervals_2 = [
            Interval(start_time, start_time + timedelta(days=3)),
            Interval(start_time + timedelta(days=3), start_time + timedelta(days=4)),
            Interval(start_time + timedelta(days=4), start_time + timedelta(days=20)),
            Interval(start_time + timedelta(days=23), start_time + timedelta(days=29)),
        ]

        missing_intervals_2 = [
            Interval(start_time + timedelta(days=20), start_time + timedelta(days=23)),
        ]

        current_intervals_3 = []

        missing_intervals_3 = [expected_interval]

        current_intervals_4 = [
            Interval(start_time, start_time + timedelta(days=3)),
            Interval(start_time + timedelta(days=3), start_time + timedelta(days=4)),
            Interval(start_time + timedelta(days=4), start_time + timedelta(days=23)),
            Interval(start_time + timedelta(days=23), start_time + timedelta(days=28)),
        ]

        missing_intervals_4 = [
            Interval(start_time + timedelta(days=28), start_time + timedelta(days=29)),
        ]

        current_intervals_5 = [
            Interval(start_time + timedelta(days=28), start_time + timedelta(days=29)),
        ]

        missing_intervals_5 = [
            Interval(start_time, start_time + timedelta(days=28)),
        ]

        self.assertEqual(
            missing_intervals_1,
            self.baseline_service._get_missing_baseline_intervals(current_intervals_1, expected_interval),
        )

        self.assertEqual(
            missing_intervals_2,
            self.baseline_service._get_missing_baseline_intervals(current_intervals_2, expected_interval),
        )

        self.assertEqual(
            missing_intervals_3,
            self.baseline_service._get_missing_baseline_intervals(current_intervals_3, expected_interval),
        )

        self.assertEqual(
            current_intervals_3,
            self.baseline_service._get_missing_baseline_intervals(missing_intervals_3, expected_interval),
        )

        self.assertEqual(
            missing_intervals_4,
            self.baseline_service._get_missing_baseline_intervals(current_intervals_4, expected_interval),
        )

        self.assertEqual(
            missing_intervals_5,
            self.baseline_service._get_missing_baseline_intervals(current_intervals_5, expected_interval),
        )

    def test_evict_off_boarded_sites(self):
        onboarded_sites = ["site1", "site2", "site3"]
        self._mock_data = Mock(Data(Mock(Minio)))
        self._mock_data.get_baseline_names_with_prefix = lambda _x: [
            "org1/vscan/site1/partial/baseline1",
            "org1/vscan/site2/partial/baseline2",
            "org1/vscan/site4/partial/baseline4",
        ]
        objects_to_remove = ["org1/vscan/site4/partial/baseline4"]
        self._mock_data.remove_baselines = Mock()
        original_data = self.baseline_service._data
        self.baseline_service._data = self._mock_data
        self.baseline_service.evict_off_boarded_sites("vscan", "org1", onboarded_sites)
        self._mock_data.remove_baselines.assert_called_once_with(objects_to_remove)
        self.baseline_service._data = original_data

    def test_get_max_baseline_duration_days(self):
        mock_read_all_partial_baselines = Mock()
        mock_read_all_partial_baselines.return_value = [
            PartialBaseline(Interval(datetime.now() - timedelta(days=5), datetime.now()), io.BytesIO()),
            PartialBaseline(Interval(datetime.now() - timedelta(days=7), datetime.now()), io.BytesIO()),
            PartialBaseline(Interval(datetime.now() - timedelta(days=2), datetime.now()), io.BytesIO()),
        ]
        original_read_all_partial_baselines = self.baseline_service.read_all_partial_baselines
        self.baseline_service.read_all_partial_baselines = mock_read_all_partial_baselines
        self.assertEqual(14, self.baseline_service.get_max_baseline_duration_days("vscan", "orgId", "site1"))
        self.baseline_service.read_all_partial_baselines = original_read_all_partial_baselines

    def test_convert_to_daily_intervals(self):
        start_time = datetime(year=2023, month=1, day=1)
        missing_intervals = [Interval(start_time + timedelta(days=2), start_time + timedelta(days=5))]
        query_intervals = [
            Interval(start_time + timedelta(days=2), start_time + timedelta(days=3)),
            Interval(start_time + timedelta(days=3), start_time + timedelta(days=4)),
            Interval(start_time + timedelta(days=4), start_time + timedelta(days=5)),
        ]
        query_intervals.sort(key=lambda i: i.end_time, reverse=True)
        self.assertEqual(query_intervals, self.baseline_service.convert_to_daily_intervals(missing_intervals))

    def test_trim_baseline_intervals(self):
        beginning_of_today = datetime(year=2023, month=1, day=25)
        lower_bound = timedelta(days=10)
        missing_intervals = [
            Interval(beginning_of_today - timedelta(days=5), beginning_of_today - timedelta(days=2)),
            Interval(beginning_of_today - timedelta(days=12), beginning_of_today - timedelta(days=6)),
        ]

        adjusted_intervals = [
            missing_intervals[0],
            Interval(beginning_of_today - lower_bound, beginning_of_today - timedelta(days=6)),
        ]

        self.assertEqual(
            adjusted_intervals,
            self.baseline_service.trim_baseline_intervals(missing_intervals, beginning_of_today, lower_bound),
        )

    def test_evict_old_partial_baselines(self):
        date1 = datetime(year=2023, month=10, day=1, hour=12, second=0)
        date2 = datetime(year=2023, month=10, day=5, hour=20, second=0)
        date3 = datetime.now() - timedelta(days=2)
        date4 = datetime.now()
        name1 = f"org-1/vscan/site-1/partial/{date1.isoformat()}_{date2.isoformat()}-baseline"
        name2 = f"org-1/vscan/site-1/partial/{date3.isoformat()}_{date4.isoformat()}-baseline"
        old_partial_baselines = [name1]

        self._mock_data = Mock(Data(Mock(Minio)))
        self._mock_data.get_baseline_names_with_prefix = lambda _x: [name1, name2]
        self._mock_data.remove_baselines = Mock()

        original_data = self.baseline_service._data
        self.baseline_service._data = self._mock_data

        self.baseline_service.evict_old_partial_baselines("vscan", "org1", timedelta(days=4))
        self._mock_data.remove_baselines.assert_called_once_with(old_partial_baselines)
        self.baseline_service._data = original_data

    def test_evict_old_combined_baselines(self):
        date1 = datetime(year=2023, month=10, day=1, hour=12, second=0)
        date2 = datetime.now() - timedelta(days=2)
        name1 = f"org-1/vscan/site-1/combined/{date1.isoformat()}-baseline"
        name2 = f"org-1/vscan/site-1/combined/{date2.isoformat()}-baseline"
        old_combined_baselines = [name1]

        self._mock_data = Mock(Data(Mock(Minio)))
        self._mock_data.get_baseline_names_with_prefix = lambda _x: [name1, name2]
        self._mock_data.remove_baselines = Mock()

        original_data = self.baseline_service._data
        self.baseline_service._data = self._mock_data

        self.baseline_service.evict_old_combined_baselines("vscan", "org1", timedelta(days=4))
        self._mock_data.remove_baselines.assert_called_once_with(old_combined_baselines)
        self.baseline_service._data = original_data
