import unittest
from dataclasses import dataclass
from datetime import datetime, timedelta
from unittest.mock import Mock

from minio import Minio

from common.utils.baseline.data import BaselineNameParseError, Data


@dataclass
class MockMinioObject:
    object_name: str
    data: bytes = b"test data"

    def close(self):
        pass

    def release_conn(self):
        pass


class TestData(unittest.TestCase):
    def setUp(self):
        self._mock_minio_client = Mock(Minio)
        self.data = Data(self._mock_minio_client)

    def test_baseline_start_time(self):
        date1 = datetime(year=2023, month=1, day=1, hour=12, second=30)
        date2 = datetime(year=2023, month=10, day=5, hour=20, second=0)
        name1 = f"org-1/vscan/site-1/partial/{date1.isoformat()}_{date2.isoformat()}-baseline"

        date3 = datetime(year=2023, month=12, day=3, hour=10, second=0)
        name2 = f"org-1/vscan/site-1/partial/{date2.isoformat()}_{date3.isoformat()}-baseline"

        name3 = "org-1/vscan/site-1/partial/baseline"
        name4 = "org-1/vscan/site-1/partial/-baseline"

        self.assertEqual(date1, Data.partial_baseline_start_time(name1))
        self.assertEqual(date2, Data.partial_baseline_start_time(name2))
        with self.assertRaises(BaselineNameParseError):
            Data.partial_baseline_start_time(name3)
        with self.assertRaises(BaselineNameParseError):
            Data.partial_baseline_start_time(name4)

    def test_baseline_end_time(self):
        date1 = datetime(year=2023, month=1, day=1, hour=12, second=30)
        date2 = datetime(year=2023, month=10, day=5, hour=20, second=0)
        name1 = f"org-1/vscan/site-1/partial/{date1.isoformat()}_{date2.isoformat()}-baseline"

        date3 = datetime(year=2023, month=12, day=3, hour=10, second=0)
        name2 = f"org-1/vscan/site-1/partial/{date2.isoformat()}_{date3.isoformat()}-baseline"

        name3 = "org-1/vscan/site-1/partial/baseline"
        name4 = "org-1/vscan/site-1/partial/-baseline"

        self.assertEqual(date2, Data.partial_baseline_end_time(name1))
        self.assertEqual(date3, Data.partial_baseline_end_time(name2))
        with self.assertRaises(BaselineNameParseError):
            Data.partial_baseline_end_time(name3)
        with self.assertRaises(BaselineNameParseError):
            Data.partial_baseline_end_time(name4)

    def test_remove_prefix_from_baseline_name(self):
        date1 = datetime(year=2023, month=1, day=1, hour=12, second=30)
        name1 = f"org-1/vscan/site-1/partial/{date1.isoformat()}-baseline"

        date2 = datetime(year=2023, month=10, day=5, hour=20, second=0)
        name2 = f"org-1/vscan/site-1/partial/{date2.isoformat()}-baseline"

        name3 = "org-1/vscan/site-1/partial/baseline"
        name4 = "org-1/vscan/site-1/partial/-baseline"
        name5 = "baseline"

        self.assertEqual(f"{date1.isoformat()}-baseline", Data.remove_prefix_from_baseline_name(name1))
        self.assertEqual(f"{date2.isoformat()}-baseline", Data.remove_prefix_from_baseline_name(name2))
        self.assertEqual("baseline", Data.remove_prefix_from_baseline_name(name3))
        self.assertEqual("-baseline", Data.remove_prefix_from_baseline_name(name4))
        with self.assertRaises(BaselineNameParseError):
            Data.remove_prefix_from_baseline_name(name5)

    def test_parse_site_id_from_baseline_name(self):
        date1 = datetime(year=2023, month=1, day=1, hour=12, second=30)
        name1 = f"org-1/vscan/site-1/partial/{date1.isoformat()}-baseline"

        date2 = datetime(year=2023, month=10, day=5, hour=20, second=0)
        name2 = f"org-1/vscan/site-2/partial/{date2.isoformat()}-baseline"

        name3 = "org-1/vscan/site-1/partial/baseline"
        name4 = "org-1/vscan/site-2/partial/-baseline"
        name5 = "baseline"

        self.assertEqual("site-1", Data.parse_site_id_from_baseline_name(name1))
        self.assertEqual("site-2", Data.parse_site_id_from_baseline_name(name2))
        self.assertEqual("site-1", Data.parse_site_id_from_baseline_name(name3))
        self.assertEqual("site-2", Data.parse_site_id_from_baseline_name(name4))
        with self.assertRaises(BaselineNameParseError):
            self.data.parse_site_id_from_baseline_name(name5)

    def test_get_partial_baselines_with_prefix(self):
        self._mock_minio_client.bucket_exists = lambda _: True
        start_time = datetime.now() - timedelta(days=1)
        end_time = datetime.now()
        self._mock_minio_client.list_objects = lambda _x, prefix, recursive: [
            MockMinioObject(f"test/prefix/test/partial/{start_time.isoformat()}_{end_time.isoformat()}-baseline")
        ]

        self._mock_minio_client.get_object = lambda _x, _y: MockMinioObject("test")

        result = self.data.get_partial_baselines_with_prefix("test/prefix/test/partial/")

        self.assertEqual(1, len(result))
        self.assertEqual(start_time, result[0].interval.start_time)
        self.assertEqual(end_time, result[0].interval.end_time)

    def test_get_combined_baseline_with_name(self):
        self._mock_minio_client.bucket_exists = lambda _: True
        time1 = datetime.now() - timedelta(days=2)
        name1 = f"org-1/vscan/site-1/combined/{time1.isoformat()}-baseline1"
        self._mock_minio_client.list_objects = lambda _x, prefix, recursive: [MockMinioObject(name1)]
        self._mock_minio_client.get_object = lambda _x, _y: MockMinioObject(name1)
        result = self.data.get_combined_baseline_with_name(name1)
        if result is None:
            self.fail("result cannot be None")
        self.assertEqual(time1, result.end_time)
