import io
import unittest
from dataclasses import dataclass
from datetime import datetime
from pickle import dump, load
from unittest.mock import Mock

from common.utils.checkpoint.data import Checkpoint, Data
from minio import Minio


@dataclass
class MockMinioObject:
    object_name: str
    data: memoryview | None

    def close(self):
        pass

    def release_conn(self):
        pass


class TestData(unittest.TestCase):
    def setUp(self):
        self._mock_minio_client = Mock(Minio)
        self.data = Data(self._mock_minio_client)

    def test_get_or_create_bucket(self):
        self._mock_minio_client.bucket_exists = lambda _: False
        self._mock_minio_client.make_bucket = Mock()

        self.data.get_or_create_bucket()
        self._mock_minio_client.make_bucket.assert_called_once()

    def test_get_checkpoint_with_prefix(self):
        self._mock_minio_client.bucket_exists = lambda _: True
        self._mock_minio_client.list_objects = lambda _x, prefix, recursive: [
            MockMinioObject("checkpoint 1", None)]

        data = io.BytesIO()
        dump("some test data", data)

        checkpoint = io.BytesIO()
        checkpoint_data = Checkpoint(datetime.now(), "1", data)
        dump(checkpoint_data, checkpoint)
        checkpoint.seek(0)
        self._mock_minio_client.get_object = lambda _x, _y: MockMinioObject(
            "checkpoint 1", checkpoint.getbuffer())

        result = self.data.get_checkpoint_with_prefix("/some/prefix")

        self.assertNotEqual(None, result)
        if result is not None:
            result.data.seek(0)
            result_data = load(result.data)
            self.assertEqual("some test data", result_data)

    def test_create_checkpoint(self):
        self._mock_minio_client.bucket_exists = lambda _: True
        self._mock_minio_client.put_object = Mock()

        data = io.BytesIO()
        dump("some test data", data)
        checkpoint = Checkpoint(datetime.now(), "1", data)
        self.data.create_checkpoint("checkpoint name", checkpoint)

        self._mock_minio_client.put_object.assert_called_once()
