import io
import time
import unittest
from datetime import datetime, timedelta
from unittest.mock import Mock

from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.common_model import DetectorType, Interval
from common.utils.detector.config import Config
from common.utils.detector.core import (
    BaselineCleanupError,
    BatchDetector,
    EventInfo,
)
from common.utils.event.core import EventService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    AnomalyExclusionSettings,
    DetectorSettings,
    UserConfigInfo,
)
from common.utils.userconfig.data import DetectorRunStatus, DetectorStatusLabel


class TestBatchDetector(BatchDetector):
    def __init__(self):
        mock_user_config_service = Mock(UserConfigService)
        user_config_infos = []
        for i in range(2):
            user_config_infos.append(
                UserConfigInfo(
                    site_id=f"site id {i}",
                    name=DetectorType.VERTICAL_PORT_SCAN.name,
                    org_id="default",
                    version="v1",
                    last_modified_time=0,
                    enabled=True,
                    silent=False,
                    source=f"source {i}",
                    block=False,
                    block_confidence_threshold=90,
                    mitre_tactics=[],
                    mitre_techniques=[],
                    settings=DetectorSettings([]),
                    exclusion_settings=AnomalyExclusionSettings([]),
                    run_status=DetectorRunStatus(DetectorStatusLabel.NOT_STARTED),
                )
            )
        mock_user_config_service.get_user_config = lambda _name, _org_id: user_config_infos
        mock_event_service = Mock(EventService)
        mock_baseline_service = Mock(BaselineService)
        mock_checkpoint_service = Mock(CheckpointService)
        mock_baseline_service.read_all_baselines = lambda _detector_name, _org_id, _site_id: []
        super().__init__(
            mock_baseline_service, mock_checkpoint_service, mock_user_config_service, mock_event_service, datetime.now()
        )

    @staticmethod
    def detector_name() -> DetectorType:
        return DetectorType.VERTICAL_PORT_SCAN

    @staticmethod
    def _baseline_max_retention_time() -> timedelta:
        return timedelta(days=30)

    def _detector_specific_baseline_cleanup(self) -> None:
        time.sleep(0.5)

    def _generate_baseline(self, site_id: str, interval: Interval) -> PartialBaseline:
        return PartialBaseline(Interval(datetime.now(), datetime.now()), io.BytesIO())

    def _combine_baseline_objects(self, baselines: dict[datetime, io.BytesIO]) -> io.BytesIO | None:
        time.sleep(0.5)
        return None

    def _detect_events(
        self, site_id: str, baseline: io.BytesIO, checkpoint: io.BytesIO
    ) -> tuple[list[EventInfo], io.BytesIO | None]:
        return [], None

    def _get_detector_run_status(
        self, status_label: DetectorStatusLabel, site_id: str, last_successful_run_time: int, msg: str = ""
    ) -> dict:
        return {}


class TestCore(unittest.TestCase):
    def setUp(self):
        self.batch_detector = TestBatchDetector()

    def test_timeout(self):
        Config.CLEANUP_TIMEOUT_SEC = "0.1"
        self.assertRaises(BaselineCleanupError, self.batch_detector.run)
        Config.CLEANUP_TIMEOUT_SEC = "10"
