import unittest
from uuid import uuid4

from confluent_kafka import Producer

from common.utils.common_model import DetectorType
from common.utils.event import data
from common.utils.event.config import Config
from vmware.nsx.intelligence.core.v1 import core_types_pb2
from vmware.nsx.intelligence.nta.v1 import nta_pb2


class TestData(unittest.TestCase):
    def setUp(self):
        self.data = data.Data(Producer({}))

        versions = ["1", "2", "3", "4", "5"]
        detector_names = [
            DetectorType.VERTICAL_PORT_SCAN,
            DetectorType.HORIZONTAL_PORT_SCAN,
            DetectorType.REMOTE_SERVICES,
            DetectorType.DATA_UPDOWNLOADER,
            DetectorType.LLMNR_NBTNS,
        ]
        detection_start_time_epoch_millis = [1677374061480, 1677374090304, 1677374103205, 1677374111607, 1677374161331]
        detection_end_time_epoch_millis = [1677374119452, 1677374125267, 1677374132089, 1677374138250, 1677374169767]
        scores = [data.Score(25, 10), data.Score(50, 30), data.Score(20, 80), data.Score(42, 12), data.Score(95, 10)]
        event_metadatas = [
            data.EventMetadata(
                {
                    "key": data.ListFlow(
                        [
                            data.Flow("src 1", "dst 1", 22, 22, data.Protocol.TCP),
                            data.Flow("src 2", "dst 2", 80, 80, data.Protocol.UDP),
                        ]
                    )
                }
            ),
            data.EventMetadata({"key": data.JsonValue("{  }")}),
            data.EventMetadata({"key": data.ListBoolean([True, False, True])}),
            data.EventMetadata({"key": data.ListString(["one", "two", "three"])}),
            data.EventMetadata({"key": data.ListNumber([1, 2, 3])}),
        ]
        merge_plans = [
            data.MergePlan("key", {"policy": data.MergePolicy.OVERRIDE_EXISTING}),
            data.MergePlan("key", {"policy": data.MergePolicy.KEEP_ORIGINAL}),
            data.MergePlan("key", {"policy": data.MergePolicy.INVALID}),
            data.MergePlan("key", {"policy": data.MergePolicy.OVERRIDE_EXISTING}),
            data.MergePlan("key", {"policy": data.MergePolicy.SET_UNION}),
            data.MergePlan("key", {"policy": data.MergePolicy.SET_UNION_WITH_MAX}),
        ]
        mitre_tactics = [
            data.Tactic.COLLECTION,
            data.Tactic.OTHER,
            data.Tactic.DISCOVERY,
            data.Tactic.CREDENTIAL_ACCESS,
            data.Tactic.LATERAL_MOVEMENT,
        ]
        mitre_techniques = [
            data.Technique.DATA_STAGED,
            data.Technique.OTHER,
            data.Technique.DOMAIN_GENERATION_ALGORITHMS,
            data.Technique.EXFILTRATION_OVER_C2_CHANNEL,
            data.Technique.MULTI_STAGE_CHANNELS,
        ]
        silents = [False, True, True, False, False]
        site_ids = [["site 1"], ["site 1", "site 2"], ["site 3"], ["site 1", "site 3"], ["site 2"]]
        org_ids = ["org 1", "org 2", "org 3", "org 4", "org 5"]

        site_id = str(uuid4())
        src_compute_id = str(uuid4())
        src_compute = data.Compute(
            data.ComputeType.COMPUTE_TYPE_VM,
            "192.168.1.1",
            src_compute_id,
            None,
            site_id,
            data.IpMask.PRIVATE,
        )
        affected_compute = src_compute

        dst_compute_id = str(uuid4())
        dst_compute = data.Compute(
            data.ComputeType.COMPUTE_TYPE_VM,
            "192.168.1.2",
            dst_compute_id,
            None,
            site_id,
            data.IpMask.PRIVATE,
        )

        data_flows = [
            data.DataFlow(
                src_compute=src_compute,
                dst_compute=dst_compute,
                transport_protocols=[data.Protocol.TCP, data.Protocol.UDP],
                application_protocols=["HTTP", "TLS"],
                server_ports=[80, 443],
                start_time=123,
                end_time=456,
            ),
        ] * (Config.MAX_DATA_FLOWS + 1)

        self.infos = []
        for i in range(5):
            self.infos.append(
                data.EventInfo(
                    version=versions[i],
                    detector_name=detector_names[i],
                    detection_start_time_epoch_milli=detection_start_time_epoch_millis[i],
                    detection_end_time_epoch_milli=detection_end_time_epoch_millis[i],
                    score=scores[i],
                    event_metadata=event_metadatas[i],
                    merge_plan=merge_plans[i],
                    mitre_tactic=mitre_tactics[i],
                    mitre_technique=mitre_techniques[i],
                    silent=silents[i],
                    site_ids=site_ids[i],
                    org_id=org_ids[i],
                    affected_compute=affected_compute,
                    is_attacker=True,
                    is_client=None,
                    data_flows=data_flows,
                )
            )

    def _compare_compute(self, value, expected):
        self.assertEqual(value.compute_type.value, expected.compute_type)
        self.assertEqual(
            data.Data._ip_address_str_to_message(value.ip_address), expected.ip_address
        )
        self.assertEqual(
            data.Data._uuid_str_to_message(value.compute_id), expected.compute_id
        )
        if value.fqdn:
            self.assertEqual(value.fqdn, expected.fqdn)
        self.assertEqual(
            data.Data._uuid_str_to_message(value.site_id), expected.site_id
        )
        self.assertEqual(value.ip_mask.value, expected.ip_mask)

    def _compare_data_flows(self, value, expected):
        self.assertEqual(len(expected), Config.MAX_DATA_FLOWS)
        for i in range(Config.MAX_DATA_FLOWS):
            data_flow = value[i]
            expected_data_flow = expected[i]
            self._compare_compute(data_flow.src_compute, expected_data_flow.src_compute)
            self._compare_compute(data_flow.dst_compute, expected_data_flow.dst_compute)
            self.assertEqual(
                [proto.value for proto in data_flow.transport_protocols],
                expected_data_flow.transport_protocols
            )
            self.assertEqual(
                data_flow.application_protocols, expected_data_flow.application_protocols
            )
            self.assertEqual(data_flow.server_ports, expected_data_flow.server_ports)
            self.assertEqual(data_flow.start_time, expected_data_flow.start_time)
            self.assertEqual(data_flow.end_time, expected_data_flow.end_time)

    def test_info_into_proto(self):
        for i in range(5):
            info: data.EventInfo = self.infos[i]
            proto: nta_pb2.Event = self.data._info_to_proto(info)

            self.assertEqual(info.version, proto.version)
            self.assertEqual(info.detector_name.name, proto.detector_name)
            self.assertEqual(info.detection_start_time_epoch_milli, proto.detection_start_time_epoch_milli)
            self.assertEqual(info.detection_end_time_epoch_milli, proto.detection_end_time_epoch_milli)
            self.assertEqual(info.score.confidence, proto.score.confidence)
            self.assertEqual(info.score.severity, proto.score.severity)

            self.assertEqual(info.merge_plan.dedup_key, proto.merge_plan.dedup_key)
            self.assertEqual(
                info.merge_plan.metadata_key_merge_operation["policy"].value,
                proto.merge_plan.metadata_key_merge_operation["policy"],
            )

            self.assertEqual(info.mitre_tactic.name, proto.mitre_tactic)
            self.assertEqual(info.mitre_technique.name, proto.mitre_technique)
            self.assertEqual(info.silent, proto.silent)
            self.assertEqual(info.site_ids, proto.site_ids)
            self.assertEqual(info.org_id, proto.org_id)

            self.assertEqual(info.is_attacker, proto.is_attacker)

            self._compare_compute(info.affected_compute, proto.affected_compute)

            self._compare_data_flows(info.data_flows, proto.data_flows)

            if i == 1:
                self.assertEqual(
                    info.event_metadata.kv_store["key"].json_value, proto.event_metadata.kv_store.get("key").json_value
                )  # type: ignore
            if i == 2:
                self.assertEqual(
                    info.event_metadata.kv_store["key"].list_bool_value,
                    proto.event_metadata.kv_store.get("key").list_bool_value.list_boolean_value,
                )  # type: ignore
            if i == 3:
                self.assertEqual(
                    info.event_metadata.kv_store["key"].list_string_value,
                    proto.event_metadata.kv_store.get("key").list_string_value.list_string_value,
                )  # type: ignore

    def test_proto_ser_de(self):
        proto = self.data._info_to_proto(self.infos[0])
        proto_ser = proto.SerializeToString()
        proto_de = nta_pb2.Event()
        proto_de.ParseFromString(proto_ser)
        self.assertEqual(proto, proto_de)
