from dataclasses import dataclass

import psycopg
from common.utils import logging
from common.utils.appinfo.config import Config
from psycopg.rows import class_row


@dataclass(frozen=True, eq=True)
class AppIdProtocol:
    """
    AppId + Protocol

    e.g.
      id  | portprotocolmappingid | displayname
    ------+-----------------------+-------------
       50 | TCP:25                | SMTP

    (e.g. SMTP: 50,TCP)
    """
    app_id: int
    protocol: str


@dataclass
class AppNameStandardPort:
    """
      id  | portprotocolmappingid | displayname
    ------+-----------------------+-------------
       56 | TCP:137               | WINS
       56 | TCP:139               | WINS
       56 | UDP:137               | WINS
       56 | UDP:138               | WINS

    (e.g. WINS: [137, 138, 139])
    """
    display_name: str
    expected_ports: list[int]


@dataclass
class AppInfoEntity:
    id: int
    protocol: str
    port: int
    displayname: str


class Data:
    logger = logging.get_common_logger(__name__)

    connect_string_format = "dbname={dbname} user={user} password={password} host={host} port={port}"

    def _entity_to_key(self, record: AppInfoEntity) -> AppIdProtocol:
        return AppIdProtocol(
            record.id,
            record.protocol
        )

    def get_app_info(self) -> dict[AppIdProtocol, AppNameStandardPort]:
        connection_string = Data.connect_string_format.format(
            dbname=Config.POSTGRES_DB_NAME,
            user=Config.POSTGRES_USER,
            password=Config.POSTGRES_PASSWORD,
            host=Config.POSTGRES_HOST,
            port=Config.POSTGRES_PORT
        )

        with psycopg.connect(connection_string) as conn:
            with conn.cursor(row_factory=class_row(AppInfoEntity)) as cur:
                query = ("select ai.id, ppm.protocol, ai.displayname, ppm.port from "
                         "appinfo ai, "
                         "appportprotocolmappingmapping appm, "
                         "portprotocolmappinginfo ppm "
                         "where ai.id=appm.appid and ppm.id = appm.portprotocolmappingid")

                cur.execute(query)

                d = dict()
                for record in cur:
                    key = self._entity_to_key(record)  # key: AppIdProtocol

                    if key not in d:
                        d[key] = AppNameStandardPort(
                            display_name=record.displayname,
                            expected_ports=[record.port]
                        )
                    else:
                        data = d[key]
                        data.expected_ports.append(record.port)

                Data.logger.info(f"retrieved appinfo records {d}")

                return d
