from datetime import date, datetime, timedelta

import urllib3
from minio import Minio

from common.utils import logging
from common.utils.baseline import const
from common.utils.baseline.config import Config
from common.utils.baseline.data import CombinedBaseline, Data, PartialBaseline
from common.utils.common_model import Interval


class BaselineError(Exception):
    pass


class BaselineService:
    logger = logging.get_common_logger(__name__)

    def __init__(self):
        self._data = Data(
            Minio(
                Config.MINIO_HOST,
                access_key=Config.MINIO_ACCESS_KEY_ID,
                secret_key=Config.MINIO_SECRET_KEY,
                http_client=urllib3.PoolManager(
                    num_pools=int(Config.MINIO_CLIENT_NUM_POOLS),
                    cert_reqs="CERT_REQUIRED",
                    ca_certs=Config.MINIO_SSL_CA_LOCATION,
                    cert_file=Config.MINIO_SSL_CERTIFICATE_LOCATION,
                    key_file=Config.MINIO_SSL_KEY_LOCATION,
                    key_password=Config.MINIO_SSL_KEY_PASSWORD,
                ),
            )
        )

    @staticmethod
    def convert_to_daily_intervals_with_trim(
        missing_intervals: list[Interval], beginning_of_today: datetime, lower_bound: timedelta
    ) -> list[Interval]:
        return BaselineService.convert_to_daily_intervals(
            BaselineService.trim_baseline_intervals(missing_intervals, beginning_of_today, lower_bound)
        )

    @staticmethod
    def trim_baseline_intervals(
        missing_intervals: list[Interval], beginning_of_today: datetime, lower_bound: timedelta
    ) -> list[Interval]:
        """Remove intervals we don't want to create baselines for"""
        adjusted_intervals = []
        max_historical = beginning_of_today - lower_bound
        # Only generate baseline for up to one week back correlated_flow does not support day granularity after one week
        for interval in missing_intervals:
            if interval.start_time > max_historical:
                # the full interval is within one week window
                adjusted_intervals.append(interval)
            elif interval.end_time > max_historical:
                # only the end time is in the one week window, adjust the start time
                adjusted_intervals.append(Interval(start_time=max_historical, end_time=interval.end_time))

        return adjusted_intervals

    @staticmethod
    def convert_to_daily_intervals(missing_intervals: list[Interval]) -> list[Interval]:
        """Breakup the intervals that were found to be missing from the baseline store into reasonable size, usually around 1 day in duration"""
        query_intervals = []
        for interval in missing_intervals:
            while interval.delta() > timedelta(days=1):
                query_intervals.append(
                    Interval(start_time=interval.start_time, end_time=interval.start_time + timedelta(days=1))
                )
                interval.start_time = interval.start_time + timedelta(days=1)
            query_intervals.append(interval)

        # prioritize generating most recent baselines first
        query_intervals.sort(key=lambda i: i.end_time, reverse=True)

        return query_intervals

    def evict_all(self, detector_name: str, org_id: str) -> None:
        try:
            prefix = f"{org_id}/{detector_name}/"
            baseline_names = self._data.get_baseline_names_with_prefix(prefix)
            BaselineService.logger.info("off-boarding baselines: %s", baseline_names)
            self._data.remove_baselines(baseline_names)

        except Exception as e:
            raise BaselineError("an error occurred evicting off-boarded baselines") from e

    def evict_off_boarded_sites(self, detector_name: str, org_id: str, on_boarded_sites: list[str]) -> None:
        try:
            prefix = f"{org_id}/{detector_name}/"
            baseline_names = self._data.get_baseline_names_with_prefix(prefix)
            name_site_pairs = [(Data.parse_site_id_from_baseline_name(name), name) for name in baseline_names]

            objects_to_remove = []
            for site_id, name in name_site_pairs:
                if site_id not in on_boarded_sites:
                    objects_to_remove.append(name)
                    BaselineService.logger.info("off-boarding baseline: %s", name)
            self._data.remove_baselines(objects_to_remove)

        except Exception as e:
            raise BaselineError("an error occurred evicting off-boarded baselines") from e

    def evict_old_partial_baselines(self, detector_name: str, org_id: str, max_retention_time: timedelta) -> None:
        try:
            prefix = f"{org_id}/{detector_name}/"
            baseline_names = self._data.get_baseline_names_with_prefix(prefix)
            partial_baseline_names = [name for name in baseline_names if Data.is_partial_baseline_name(name)]

            old_partial_baselines = []
            now = datetime.now()
            for partial_baseline_name in partial_baseline_names:
                if now - Data.partial_baseline_end_time(partial_baseline_name) > max_retention_time:
                    old_partial_baselines.append(partial_baseline_name)
                    BaselineService.logger.info(f"outdated partial baseline {partial_baseline_name}, set to be removed")

            self._data.remove_baselines(old_partial_baselines)

        except Exception as e:
            raise BaselineError("an error occurred deleting outdated partial baselines") from e

    def evict_old_combined_baselines(self, detector_name: str, org_id: str, max_retention_time: timedelta) -> None:
        try:
            prefix = f"{org_id}/{detector_name}/"
            baseline_names = self._data.get_baseline_names_with_prefix(prefix)
            combined_baseline_names = [name for name in baseline_names if Data.is_combined_baseline_name(name)]

            old_combined_baselines = []
            now = datetime.now()
            for combined_baseline_name in combined_baseline_names:
                if now - Data.combined_baseline_end_time(combined_baseline_name) > max_retention_time:
                    old_combined_baselines.append(combined_baseline_name)
                    BaselineService.logger.info(f"outdated baseline {combined_baseline_name}, set to be removed")

            self._data.remove_baselines(old_combined_baselines)

        except Exception as e:
            raise BaselineError("an error occurred deleting outdated combined baselines") from e

    def write_partial_baseline(
        self, detector_name: str, org_id: str, site_id: str, partial_baseline: PartialBaseline
    ) -> None:
        self.write_partial_baselines(detector_name, org_id, site_id, [partial_baseline])

    def write_partial_baselines(
        self, detector_name: str, org_id: str, site_id: str, partial_baselines: list[PartialBaseline]
    ) -> None:
        try:
            prefix = f"{org_id}/{detector_name}/{site_id}/{const.PARTIAL}/"
            self._data.create_partial_baselines(prefix, partial_baselines)
        except Exception as e:
            raise BaselineError("an error occurred while writing partial baselines") from e

    def read_all_partial_baselines(self, detector_name: str, org_id: str, site_id: str) -> list[PartialBaseline]:
        try:
            prefix = f"{org_id}/{detector_name}/{site_id}/{const.PARTIAL}/"
            return self._data.get_partial_baselines_with_prefix(prefix)
        except Exception as e:
            raise BaselineError("an error occurred while reading baseline from minio") from e

    def write_combined_baseline(
        self, detector_name: str, org_id: str, site_id: str, combined_baseline: CombinedBaseline
    ) -> None:
        try:
            prefix = f"{org_id}/{detector_name}/{site_id}/{const.COMBINED}/"
            self._data.create_combined_baseline(prefix, combined_baseline)
        except Exception as e:
            raise BaselineError("an error occurred while writing combined baseline") from e

    def read_combined_baseline(
        self, detector_name: str, org_id: str, site_id: str, end_time: datetime
    ) -> CombinedBaseline | None:
        try:
            prefix = f"{org_id}/{detector_name}/{site_id}/{const.COMBINED}/"
            name = f"{prefix}{end_time.isoformat()}-baseline"
            return self._data.get_combined_baseline_with_name(name)
        except Exception as e:
            raise BaselineError("an error occurred while writing combined baseline") from e

    def get_missing_baseline_intervals(
        self, detector_name: str, org_id: str, site_id: str, max_retention_time: timedelta, end_time: datetime
    ) -> list[Interval]:
        """Returns a list of intervals that represent time ranges not covered by the baselines in the baseline store

        The max interval returned is today.min.time() - max retention time, today.min.time(). In other words, the current day is never included
        in the missing interval that is returned.
        """
        if max_retention_time.days < 1:
            return []

        # create interval that we expect to be covered by the baseline
        # start_time is determined by the CURRENT REAL DAY - max retention time
        #     such that if the detector time travels we don't start to try to create partial baselines for intervals we already evicted
        # end_time is passed in to allow for detector time travel
        expected_interval = Interval(
            start_time=datetime.combine(date.today(), datetime.min.time()) - max_retention_time, end_time=end_time
        )

        # get list of intervals currently in the store
        prefix = f"{org_id}/{detector_name}/{site_id}/{const.PARTIAL}/"
        current_intervals = [
            Interval(Data.partial_baseline_start_time(name), Data.partial_baseline_end_time(name))
            for name in self._data.get_baseline_names_with_prefix(prefix)
        ]
        if current_intervals == []:
            return [expected_interval]
        return self._get_missing_baseline_intervals(current_intervals, expected_interval)

    def _get_missing_baseline_intervals(
        self, current_intervals: list[Interval], expected_interval: Interval
    ) -> list[Interval]:
        """Returns a list of timestamps, each represents an object missing from the baseline store

        Baseline objects in the baseline store are chunked into time periods, for example:
        2023-01-01T00:00:00_2023-01-02T00:00:00-baseline
        2023-01-02T00:00:00_2023-01-03T00:00:00-baseline
        2023-01-03T00:00:00_2023-01-04T00:00:00-baseline
        ...

        If one is missing, for example:
        2023-01-01T00:00:00_2023-01-02T00:00:00-baseline
        (expecting day 2 here)
        2023-01-03T00:00:00_2023-01-04T00:00:00-baseline

        Or more likely, when a day has passed:
        2023-01-01T00:00:00_2023-01-02T00:00:00-baseline
        2023-01-02T00:00:00_2023-01-03T00:00:00-baseline
        2023-01-03T00:00:00_2023-01-04T00:00:00-baseline
        ... (expecting day 4 here)

        We need to detect this and let the detector know that it needs to generate these missing baselines

        Missing interval procedure example

        expected_interval = 1-29
        current_intervals = 2-3, 3-4, 23-25, 4-20  # need to identify missing intervals 1-2, 20-23, 25-29

        1. sort current intervals
        2-3, 3-4, 4-20, 23-25

        2. identify bookends
        1-2, 25-29

        3. identify gaps
        20-23
        """
        if current_intervals == []:
            return [expected_interval]

        # assumes no overlapping intervals and i1.end_time == i2.start_time means we can combine time intervals
        current_intervals.sort(key=lambda i: i.start_time)

        missing_intervals = []
        if expected_interval.start_time < current_intervals[0].start_time:
            missing_intervals.append(
                Interval(start_time=expected_interval.start_time, end_time=current_intervals[0].start_time)
            )

        for i in range(len(current_intervals) - 1):
            start_time = current_intervals[i].end_time
            end_time = current_intervals[i + 1].start_time
            if start_time != end_time:
                missing_intervals.append(Interval(start_time, end_time))

        if expected_interval.end_time > current_intervals[-1].end_time:
            missing_intervals.append(
                Interval(start_time=current_intervals[-1].end_time, end_time=expected_interval.end_time)
            )

        return missing_intervals

    def get_max_baseline_duration_days(self, detector_name: str, org_id: str, site_id: str) -> int:
        try:
            partial_baselines = self.read_all_partial_baselines(detector_name, org_id, site_id)
            number_of_days_with_baseline = sum([baseline.interval.delta().days for baseline in partial_baselines])
            return number_of_days_with_baseline
        except Exception as e:
            raise BaselineError("An error occurred while getting max baseline duration days") from e
