import io
from dataclasses import dataclass
from datetime import datetime

from common.utils import logging
from common.utils.baseline import const
from common.utils.baseline.config import Config
from common.utils.common_model import Interval
from dateutil import parser as dateparser
from minio import Minio


class BaselineNameParseError(Exception):
    pass


@dataclass
class PartialBaseline:
    interval: Interval
    data: io.BytesIO


@dataclass
class CombinedBaseline:
    end_time: datetime
    data: io.BytesIO


class Data:
    logger = logging.get_common_logger(__name__)

    def __init__(self, minio_client: Minio):
        self._client = minio_client

    def get_baseline_names_with_prefix(self, prefix: str) -> list[str]:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                "when trying to get partial baseline names, the baseline bucket was not found")
            return []

        objects = self._client.list_objects(
            Config.BASELINE_BUCKET, prefix=prefix, recursive=True)
        return [obj.object_name for obj in objects]

    def get_partial_baselines_with_prefix(self, prefix: str) -> list[PartialBaseline]:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                "when trying to get baselines, the baseline bucket was not found")
            return []

        object_names = self.get_baseline_names_with_prefix(prefix)
        response = None
        try:
            baselines = []
            for name in object_names:
                response = self._client.get_object(
                    Config.BASELINE_BUCKET, name)

                if response is not None:
                    baselines.append(
                        PartialBaseline(
                            interval=Interval(
                                Data.partial_baseline_start_time(name),
                                Data.partial_baseline_end_time(name)
                            ),
                            data=io.BytesIO(response.data)
                        )
                    )
            return baselines
        finally:
            if response is not None:
                response.close()
                response.release_conn()

    def get_combined_baseline_with_name(self, name: str) -> CombinedBaseline | None:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                f"when trying to get combined baseline with name {name}, the baseline bucket was not found")
            return None

        object_names = self.get_baseline_names_with_prefix(name)
        if len(object_names) == 0:
            Data.logger.warn(
                f"could not find combined baseline with name {name}")
            return None

        response = None
        try:
            response = self._client.get_object(
                Config.BASELINE_BUCKET, name)
            if response is None:
                Data.logger.warn(
                    f"got None when retrieving combined baseline {name}")
                return None
            return CombinedBaseline(
                end_time=Data.combined_baseline_end_time(name),
                data=io.BytesIO(response.data)
            )
        finally:
            if response is not None:
                response.close()
                response.release_conn()

    def create_partial_baselines(self, prefix: str, partial_baselines: list[PartialBaseline]) -> None:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.info(
                "baseline bucket not found, creating new baseline bucket")
            self._client.make_bucket(Config.BASELINE_BUCKET)

        for partial_baseline in partial_baselines:
            partial_baseline_name = f"{prefix}{partial_baseline.interval.start_time.isoformat()}_{partial_baseline.interval.end_time.isoformat()}-baseline"
            # minio client uses read() to get the bytes to write, so need to set cursor to beginning of the buffer
            partial_baseline.data.seek(0)
            self._client.put_object(
                Config.BASELINE_BUCKET,
                object_name=partial_baseline_name,
                data=partial_baseline.data,
                length=partial_baseline.data.getbuffer().nbytes
            )
            # rewind after minio write so the baseline can be read again
            partial_baseline.data.seek(0)
            Data.logger.info(
                f"created partial baseline with name: {partial_baseline_name}")

    def create_combined_baseline(self, prefix: str, baseline: CombinedBaseline) -> None:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.info(
                "baseline bucket not found, creating new baseline bucket")
            self._client.make_bucket(Config.BASELINE_BUCKET)

        baseline_name = f"{prefix}{baseline.end_time.isoformat()}-baseline"
        baseline.data.seek(0)
        # minio client uses read() to get the bytes to write, so need to set cursor to beginning of the buffer
        self._client.put_object(
            Config.BASELINE_BUCKET,
            object_name=baseline_name,
            data=baseline.data,
            length=baseline.data.getbuffer().nbytes
        )
        # rewind after minio write so the baseline can be read again
        baseline.data.seek(0)
        Data.logger.info(
            f"created baseline with name: {baseline_name}")

    def remove_baselines(self, baseline_names: list[str]) -> None:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                "when trying to remove baselines, the baseline bucket was not found")
            return

        for name in baseline_names:
            self._client.remove_object(Config.BASELINE_BUCKET, name)
            Data.logger.info(f"deleted baseline: {name}")

    @staticmethod
    def partial_baseline_start_time(partial_baseline_name: str) -> datetime:
        try:
            return dateparser.parse(Data._partial_baseline_start_and_end_time_strings(partial_baseline_name)[0])
        except dateparser.ParserError as e:
            raise BaselineNameParseError(
                f"unable to parse timestamp from name {partial_baseline_name}") from e

    @staticmethod
    def partial_baseline_end_time(partial_baseline_name: str) -> datetime:
        try:
            return dateparser.parse(Data._partial_baseline_start_and_end_time_strings(partial_baseline_name)[1])
        except dateparser.ParserError as e:
            raise BaselineNameParseError(
                f"unable to parse timestamp from name {partial_baseline_name}") from e

    @staticmethod
    def _partial_baseline_start_and_end_time_strings(partial_baseline_name: str) -> list[str]:
        name_without_prefix = Data.remove_prefix_from_baseline_name(
            partial_baseline_name)
        if "-" not in name_without_prefix:
            raise BaselineNameParseError(
                f"unable to parse timestamp from name {partial_baseline_name}, delimiter '-' not found")
        start_and_end_times = name_without_prefix[:name_without_prefix.rindex(
            "-")].split("_")
        if len(start_and_end_times) != 2:
            raise BaselineNameParseError(
                f"unable to split times from {partial_baseline_name}, got {start_and_end_times}")
        return start_and_end_times

    @staticmethod
    def combined_baseline_end_time(baseline_name: str) -> datetime:
        name_without_prefix = Data.remove_prefix_from_baseline_name(
            baseline_name)
        if "-" not in name_without_prefix:
            raise BaselineNameParseError(
                f"unable to parse timestamp from name {baseline_name}, delimiter '-' not found")
        end_time = name_without_prefix[:name_without_prefix.rindex("-")]
        try:
            return dateparser.parse(end_time)
        except dateparser.ParserError as e:
            raise BaselineNameParseError(
                f"unable to parse timestamp from name {baseline_name}") from e

    @staticmethod
    def remove_prefix_from_baseline_name(baseline_name: str) -> str:
        # <org-id>/<detector-name>/<site-id>/<type>/<timestamp>-baseline
        # return <timestamp>-baseline from the above
        name_split = baseline_name.split("/")
        if len(name_split) < 5:
            raise BaselineNameParseError(
                f"baseline name {baseline_name} in incorrect format")
        return name_split[4]

    @staticmethod
    def parse_site_id_from_baseline_name(baseline_name: str) -> str:
        # <org-id>/<detector-name>/<site-id>/<type>/<timestamp>-baseline
        # extract site-id from the above string (baseline_name)
        name_split = baseline_name.split("/")
        if len(name_split) < 5:
            raise BaselineNameParseError(
                f"baseline name {baseline_name} in incorrect format")

        return name_split[2]

    @staticmethod
    def is_partial_baseline_name(baseline_name: str) -> bool:
        # <org-id>/<detector-name>/<site-id>/<type>/<timestamp>-baseline
        # check that <type> == const.PARTIAL
        name_split = baseline_name.split("/")
        if len(name_split) < 5:
            raise BaselineNameParseError(
                f"baseline name {baseline_name} in incorrect format")

        return name_split[3] == const.PARTIAL

    @staticmethod
    def is_combined_baseline_name(baseline_name: str) -> bool:
        # <org-id>/<detector-name>/<site-id>/<type>/<timestamp>-baseline
        # check that <type> == const.COMBINED
        name_split = baseline_name.split("/")
        if len(name_split) < 5:
            raise BaselineNameParseError(
                f"baseline name {baseline_name} in incorrect format")

        return name_split[3] == const.COMBINED
