import io

import urllib3
from common.utils import logging
from common.utils.checkpoint.config import Config
from common.utils.checkpoint.data import Checkpoint, Data
from minio import Minio


class CheckpointService:
    logger = logging.get_common_logger(__name__)

    def __init__(self):
        self._data = Data(
            Minio(
                Config.MINIO_HOST,
                access_key=Config.MINIO_ACCESS_KEY_ID,
                secret_key=Config.MINIO_SECRET_KEY,
                http_client=urllib3.PoolManager(
                    num_pools=int(Config.MINIO_CLIENT_NUM_POOLS),
                    cert_reqs="CERT_REQUIRED",
                    ca_certs=Config.MINIO_SSL_CA_LOCATION,
                    cert_file=Config.MINIO_SSL_CERTIFICATE_LOCATION,
                    key_file=Config.MINIO_SSL_KEY_LOCATION,
                    key_password=Config.MINIO_SSL_KEY_PASSWORD,
                ),
            )
        )

    @staticmethod
    def get_prefix(detector_name: str, org_id: str, site_id: str) -> str:
        return f"{org_id}/{detector_name}/{site_id}/checkpoint"

    def read_checkpoint(
        self, detector_name: str, org_id: str, site_id: str
    ) -> Checkpoint | None:
        prefix = CheckpointService.get_prefix(detector_name, org_id, site_id)

        # check if the bucket + prefix exist
        # return None is the checkpoint is missing
        return self._data.get_checkpoint_with_prefix(prefix)

    def update(
        self,
        detector_name: str,
        org_id: str,
        site_id: str,
        checkpoint_buffer: io.BytesIO,
    ) -> bool:
        if checkpoint_buffer is None:
            CheckpointService.logger.info("new checkpoint is none. skip updating")
            return False

        prefix = CheckpointService.get_prefix(detector_name, org_id, site_id)
        return self._data.update(prefix, checkpoint_buffer)

    def put_object(
        self,
        bucket_name: str,
        detector_name: str,
        org_id: str,
        site_id: str,
        checkpoint_buffer: io.BytesIO,
    ) -> bool:
        prefix = CheckpointService.get_prefix(detector_name, org_id, site_id)
        return self._put_object(bucket_name, prefix, checkpoint_buffer)

    def _put_object(
        self, bucket_name: str, prefix: str, checkpoint_buffer: io.BytesIO
    ) -> bool:
        return self._data.put_object(bucket_name, prefix, checkpoint_buffer)

    def get_or_create_bucket(self) -> str:
        return self._data.get_or_create_bucket()
