import io
from dataclasses import dataclass
from datetime import datetime
from pickle import dump, load

from common.utils import logging
from common.utils.checkpoint.config import Config
from minio import Minio


@dataclass
class Checkpoint:
    time: datetime
    version: str
    data: io.BytesIO  # dataframe in parquet


class Data:
    logger = logging.get_common_logger(__name__)

    def __init__(self, minio_client: Minio):
        self._client = minio_client

    def get_checkpoint_name_with_prefix(self, prefix: str) -> str:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                "when trying to get checkpoint names, the baseline bucket was not found. Creating bucket")
            self._client.make_bucket(Config.BASELINE_BUCKET)

        objects = self._client.list_objects(
            Config.BASELINE_BUCKET, prefix=prefix, recursive=True)

        object_names = [obj.object_name for obj in objects]
        if len(object_names) == 0:
            Data.logger.info(f"{prefix} not found in {Config.BASELINE_BUCKET}")

        if len(object_names) > 1:
            # there should only be 1 file with this name, do better error check
            Data.logger.warn(
                f"Multiple {prefix} detected in {Config.BASELINE_BUCKET}")

        return next(iter(object_names), "")  # return 1st name or empty string

    def get_or_create_bucket(self) -> str:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            self._client.make_bucket(Config.BASELINE_BUCKET)

        return Config.BASELINE_BUCKET

    def get_checkpoint_with_prefix(self, prefix: str) -> Checkpoint | None:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            self._client.make_bucket(Config.BASELINE_BUCKET)
            return None  # no checkpoint since bucket is empty

        object_name = self.get_checkpoint_name_with_prefix(prefix)

        if not object_name:
            Data.logger.info(
                "Checkpoint is empty, delegate to detector to create")
            return None

        response = None
        try:
            response = self._client.get_object(
                Config.BASELINE_BUCKET, object_name)

            checkpoint_data = io.BytesIO(response.data)
            checkpoint = load(checkpoint_data)

            return checkpoint
        finally:
            if response is not None:
                response.close()
                response.release_conn()

    def put_object(self, bucket_name: str, prefix: str, checkpoint_buffer: io.BytesIO) -> bool:
        response = self._client.put_object(
            bucket_name, object_name=prefix, data=checkpoint_buffer,
            length=checkpoint_buffer.getbuffer().nbytes)

        Data.logger.info(f"checkpoint put_object: {response}")
        return True

    def update(self, prefix, checkpoint_buffer: io.BytesIO) -> bool:
        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                "when trying to update checkpoint, the baseline bucket was not found. Creating bucket")
            self._client.make_bucket(Config.BASELINE_BUCKET)

        return self.put_object(Config.BASELINE_BUCKET, prefix, checkpoint_buffer)

    def create_checkpoint(self, checkpoint_name: str, checkpoint: Checkpoint) -> None:
        # create checkpoint file in minio

        if not self._client.bucket_exists(Config.BASELINE_BUCKET):
            Data.logger.warn(
                "when trying to get checkpoint names, the baseline bucket was not found")
            return  # TODO: better error

        # minio client uses read() to get the bytes to write, so need to set cursor to beginning of the buffer
        checkpoint.data.seek(0)

        # turn checkpoint into a bytesIO
        checkpoint_buffer = io.BytesIO()
        dump(checkpoint, checkpoint_buffer)
        checkpoint_buffer.seek(0)

        self._client.put_object(
            Config.BASELINE_BUCKET, object_name=checkpoint_name, data=checkpoint_buffer,
            length=checkpoint_buffer.getbuffer().nbytes)

        Data.logger.info(
            f"created checkpoint with name: {checkpoint_name}")

    def delete_checkpoint(self, checkpoint_name: str) -> None:
        # TODO: delete the checkpoint file in minio
        pass
