import io
import sys
import traceback
from abc import ABC, abstractmethod
from concurrent.futures import Executor, ThreadPoolExecutor
from datetime import datetime, timedelta
from threading import Lock
from typing import Any, Generator

from common.utils import logging
from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import CombinedBaseline, PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.common_model import DetectorType, Interval
from common.utils.detector.config import Config
from common.utils.event.core import EventService
from common.utils.event.data import EventInfo
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import DetectorRunStatus, DetectorStatusLabel, UserConfigInfo


class BaselineCleanupError(Exception):
    pass


class BatchDetector(ABC):
    """Abstract Base Class for NTA Batch Detectors

    All NTA Detectors that work via batching should implement this base class.
    This base class is responsible for defining the shared logic between each batch detector,
    most importantly the run() method that defines the core logic of a batch detector.
    """

    detector_run_lock = Lock()
    logger = logging.get_common_logger(__name__)

    def __init__(
        self,
        baseline_service: BaselineService,
        checkpoint_service: CheckpointService,
        user_config_service: UserConfigService,
        event_service: EventService,
        now: datetime,
    ):
        self._baseline_service = baseline_service
        self._user_config_service = user_config_service
        self._event_service = event_service
        self._checkpoint_service = checkpoint_service
        self._now = now
        self._beginning_of_today = datetime.combine(now.date(), datetime.min.time())

    def run(self) -> None:
        """Manage cleanup of the threadpool and ensure only one detection is running at a time"""
        if BatchDetector.detector_run_lock.acquire(blocking=False):
            try:
                status_results: list[tuple[str, DetectorRunStatus]] = []
                with ThreadPoolExecutor(max_workers=1) as executor:
                    status_results = self._run(executor)
                BatchDetector.logger.info(f"setting status for {len(status_results)} sites")
                for site_id, status in status_results:
                    # update status for each site individually
                    BatchDetector.logger.debug(f"setting status for site {site_id}")
                    self._user_config_service.update_detector_status(
                        status, self.detector_name().name, Config.ORG_ID, site_id
                    )
            except Exception as e:
                # something bad happened, need to set failure status for all sites
                BatchDetector.logger.warn(
                    f"attempting to set status to {DetectorStatusLabel.FAILURE} for all sites for detector {self.detector_name().name} due to error {e}"
                )
                detector_status = DetectorRunStatus(DetectorStatusLabel.FAILURE, message=str(e))
                self._user_config_service.update_detector_status_all_sites(
                    detector_status, self.detector_name().name, Config.ORG_ID
                )
                raise e
            finally:
                BatchDetector.detector_run_lock.release()
        else:
            BatchDetector.logger.warning("failed to start the detector, a detector is already running")

    def _run(self, executor: Executor) -> list[tuple[str, DetectorRunStatus]]:
        """The main operating logic of an NTA batch detector

        This method is called when the detector job is started, it should perform the following:
        1. Retrieve user-defined configuration
        2. Manage the detector baseline
            a. Read in the current user configuration from the user config store
            b. Remove any baseline objects belonging to off-boarded sites
            c. Remove any baseline objects that fall outside of the detector's retention window
            d. Detect what baseline objects are missing from the baseline store
            e. Generate and write any missing baselines to the baseline store
        3. get (or create) checkpoint file
        4. Detect and report events
            a. Read in and construct the baseline from the baseline store
            b. Perform detection against the baseline, given the relevant user config
            c. Report any detected events to the event queue
        """
        user_configs_for_each_site = self._user_config_service.get_user_config(self.detector_name().name, Config.ORG_ID)
        sites = [config.site_id for config in user_configs_for_each_site]
        BatchDetector.logger.info(f"got on-boarded sites as: {sites}")

        try:
            executor.submit(self._baseline_cleanup, sites).result(timeout=float(Config.CLEANUP_TIMEOUT_SEC))
        except Exception as e:
            raise BaselineCleanupError("Failed to evict baseline objects") from e
        BatchDetector.logger.info("finished cleaning up baseline objects")

        run_status_data: list[tuple[str, DetectorRunStatus]] = []  # used to collect the run status for each site
        for config in user_configs_for_each_site:
            if not config.enabled:
                BatchDetector.logger.info(
                    f"{config.name} is not enabled for {config.site_id}, skipping baseline update and detection"
                )
                continue
            try:
                executor.submit(self._run_on, config).result(timeout=float(Config.DETECTOR_TIMEOUT_SEC))
                run_status_data.append(
                    (
                        config.site_id,
                        self._get_detector_run_status(
                            DetectorStatusLabel.SUCCESS,
                            config.site_id,
                            last_successful_run_time=int(self._now.timestamp() * 1000),
                        ),
                    )
                )
            except TimeoutError:
                BatchDetector.logger.error(
                    f"Timed out while detecting events for site {config.site_id}, some events might not be reported"
                )
                run_status_data.append(
                    (
                        config.site_id,
                        DetectorRunStatus(DetectorStatusLabel.FAILURE, message="timeout while detecting events"),
                    )
                )
            except Exception as e:
                BatchDetector.logger.error(f"Failed to detect events on site {config.site_id} due to {e}")
                BatchDetector.logger.error(traceback.format_exc())
                run_status_data.append((config.site_id, DetectorRunStatus(DetectorStatusLabel.FAILURE, message=str(e))))

        return run_status_data

    def _baseline_cleanup(self, on_boarded_sites: list[str]) -> None:
        """Perform eviction of baseline objects from the store"""
        if Config.EVICT_ALL_BASELINES == "True":
            BatchDetector.logger.warn("evicting all baselines due to config EVICT_ALL_BASELINES being set to True")
            self._baseline_service.evict_all(self.detector_name().name, Config.ORG_ID)
            return

        self._baseline_service.evict_off_boarded_sites(self.detector_name().name, Config.ORG_ID, on_boarded_sites)
        self._baseline_service.evict_old_partial_baselines(
            self.detector_name().name, Config.ORG_ID, self._baseline_max_retention_time()
        )

        # need to evict old combined baselines, but keep past two days around for historical detection runs
        self._baseline_service.evict_old_combined_baselines(self.detector_name().name, Config.ORG_ID, timedelta(days=2))

        self._detector_specific_baseline_cleanup()

    def _run_on(self, config: UserConfigInfo) -> None:
        """Update the baseline, read the baseline, prepare checkpoint detect events, and publish events"""
        # try to update partial baselines, if any update, combine all partial baselines and write the combined baseline to the baseline store
        BatchDetector.logger.info(f"running detection on site {config.site_id}")
        baseline_was_updated = self._update_baseline(config.site_id)
        baseline = self._read_full_baseline(config.site_id, baseline_was_updated)
        if baseline is None:
            BatchDetector.logger.warning(f"cannot perform detection, got None baseline for site {config.site_id}")
            return
        if baseline_was_updated:
            # write combined baseline to the baseline-store
            self._baseline_service.write_combined_baseline(
                self.detector_name().name,
                Config.ORG_ID,
                config.site_id,
                CombinedBaseline(self._beginning_of_today, baseline),
            )

        # get (or create) checkpoint
        checkpoint = self._read_checkpoint(config)
        if checkpoint is None:
            checkpoint = self._create_checkpoint(config)

        # detect & push event
        num_events_published = 0
        for event in self._detect_events(config, baseline, checkpoint):
            if num_events_published > int(Config.MAX_EVENTS_PER_RUN):
                BatchDetector.logger.warn(
                    "exceeded max number of events publishable in a single run ({}), stopping event publishing".format(
                        Config.MAX_EVENTS_PER_RUN
                    )
                )
                break
            BatchDetector.logger.info(f"Publishing event for site {config.site_id}")
            self._event_service.publish_event(event)
            num_events_published += 1

        # update checkpoint
        new_checkpoint = self._update_checkpoint(config)
        if new_checkpoint is not None:
            BatchDetector.logger.info(f"updating checkpoint({new_checkpoint}) events on site {config.site_id}")
            self._checkpoint_service.update(self.detector_name().name, Config.ORG_ID, config.site_id, new_checkpoint)

    def _update_baseline(self, site_id: str) -> bool:
        """Updates the partial baseline objects in the baseline store

        Returns True if any updates were made, False otherwise
        """
        missing_baseline_intervals = self._baseline_service.get_missing_baseline_intervals(
            self.detector_name().name,
            Config.ORG_ID,
            site_id,
            self._baseline_max_retention_time(),
            self._beginning_of_today,
        )
        if len(missing_baseline_intervals) == 0:
            BatchDetector.logger.info("No missing baseline intervals found, skipping baseline update")
            return False

        baseline_intervals = self._get_baseline_intervals(missing_baseline_intervals)

        if len(baseline_intervals) == 0:
            BatchDetector.logger.info("No new partial baselines to create, skipping baseline update")
            return False

        for interval in baseline_intervals:
            partial_baseline = self._generate_baseline(site_id, interval)
            self._baseline_service.write_partial_baselines(
                self.detector_name().name, Config.ORG_ID, site_id, [partial_baseline]
            )
        return True

    def _read_full_baseline(self, site_id: str, baseline_was_updated: bool) -> io.BytesIO | None:
        """Reads and returns the current state of the baseline

        If any partial baselines have been updated during this run, then it reads all of
        the partial baselines into memory and combines them in memory, then returns them

        If the baselines is unchanged this run, it returns the already saved combined baseline
        """
        if baseline_was_updated:
            # since baseline was updated, need to read all partial baselines in and combine in memory
            return self._read_and_combine_partial_baselines(site_id)

        # since the baseline remained the same, we can simply read in the already combined baseline in from the baseline store
        combined_baseline = self._baseline_service.read_combined_baseline(
            self.detector_name().name, Config.ORG_ID, site_id, self._beginning_of_today
        )
        if combined_baseline is None:
            BatchDetector.logger.warn(
                "tried to read combined baseline but could not find it in the baseline store, will try to combine partial baselines instead"
            )
            return self._read_and_combine_partial_baselines(site_id)

        return combined_baseline.data

    def _read_and_combine_partial_baselines(self, site_id: str) -> io.BytesIO | None:
        baselines = self._baseline_service.read_all_partial_baselines(self.detector_name().name, Config.ORG_ID, site_id)
        BatchDetector.logger.info(f"got {len(baselines)} baseline objects from the baseline store")
        self._log_size_of_baselines(baselines)
        return self._combine_baseline_objects(baselines)

    def _log_size_of_baselines(self, baselines: list[PartialBaseline]) -> None:
        size_in_MB = 0
        for baseline in baselines:
            size_in_MB += sys.getsizeof(baseline.data) / 1_000
        BatchDetector.logger.info(f"Got size of baselines to be {size_in_MB}KB")

    @staticmethod
    @abstractmethod
    def detector_name() -> DetectorType:
        """The name of the detector, should be unique among all detectors"""

    @staticmethod
    @abstractmethod
    def _baseline_max_retention_time() -> timedelta:
        """The maximum time to retain a baseline object in the baseline store"""

    def _detector_specific_baseline_cleanup(self) -> None:
        """Any cleanup / eviction of the baseline store that the detector needs to do outside of the standard"""
        return

    @abstractmethod
    def _generate_baseline(self, site_id: str, interval: Interval) -> PartialBaseline:
        """Generate a partial baseline to write to the store"""

    @abstractmethod
    def _combine_baseline_objects(self, baselines: list[PartialBaseline]) -> io.BytesIO | None:
        """Given a list of partial baselines, combine them into one baseline"""

    @abstractmethod
    def _detect_events(
        self, config: UserConfigInfo, baseline: io.BytesIO, checkpoint: io.BytesIO | None
    ) -> Generator[EventInfo, Any, Any]:
        """Given the baseline, detect events"""

    def _read_checkpoint(self, config: UserConfigInfo) -> io.BytesIO | None:
        """Given the config, read checkpoint, default implementation return None"""
        return None

    def _create_checkpoint(self, config: UserConfigInfo) -> io.BytesIO | None:
        """Given the config, create the initial checkpoint, default implementation return None"""
        return None

    def _update_checkpoint(self, user_config: UserConfigInfo) -> io.BytesIO | None:
        """Assuming the checkpoint already exists, update it

        Return None if checkpoint is not supported or no update is needed"""
        return None

    def _get_baseline_intervals(self, missing_intervals: list[Interval]) -> list[Interval]:
        """Given a list of intervals missing from the baseline store,
        return the intervals which should be used to create partial baselines

        The following is a default implementation, it's intended to be overwritten by
        the child class
        """
        return missing_intervals

    @abstractmethod
    def _get_detector_run_status(
        self, status_label: DetectorStatusLabel, site_id: str, last_successful_run_time: int, message: str = ""
    ) -> DetectorRunStatus:
        "Returns status of detector determining if and when it last successfully finished running"
