from dataclasses import dataclass
from enum import Enum, auto
from ipaddress import ip_address
from typing import cast
from uuid import UUID

from confluent_kafka import KafkaError, Message, Producer

from common.utils import logging
from common.utils.common_model import DetectorType
from common.utils.event.config import Config
from vmware.nsx.intelligence.core.v1 import core_types_pb2
from vmware.nsx.intelligence.nta.v1 import nta_core_types_pb2, nta_pb2


class UnionValueTypeError(Exception):
    pass


class EventKey:
    EVENT_METADATA_REPORT_KEY = "REPORT"
    EVENT_METADATA_VIZ_FILTER_KEY = "VIZ_FILTER"
    EVENT_METADATA_DESCRIPTION_KEY = "event_description"


class ComputeType(Enum):
    COMPUTE_TYPE_INVALID = 0
    COMPUTE_TYPE_VM = auto()
    COMPUTE_TYPE_IP = auto()
    COMPUTE_TYPE_PHYSICAL_SERVER = auto()


class IpMask(Enum):
    INVALID = 0
    PUBLIC = auto()
    PRIVATE = auto()


class Tactic(Enum):
    PERSISTENCE = 0
    CREDENTIAL_ACCESS = auto()
    DISCOVERY = auto()
    COMMAND_AND_CONTROL = auto()
    LATERAL_MOVEMENT = auto()
    COLLECTION = auto()
    EXFILTRATION = auto()
    OTHER = auto()


class Technique(Enum):
    NETWORK_SERVICE_SCANNING = 0
    NON_STANDARD_PORT = auto()
    LLMNR_NBTNS_POISONING = auto()
    REMOTE_SERVICES = auto()
    PROTOCOL_TUNNELING = auto()
    DOMAIN_GENERATION_ALGORITHMS = auto()
    EXFILTRATION_OVER_C2_CHANNEL = auto()
    EXFILTRATION_OVER_ALTERNATIVE_PROTOCOL = auto()
    DATA_STAGED = auto()
    MULTI_STAGE_CHANNELS = auto()
    OTHER = auto()


class Protocol(Enum):
    INVALID = 0
    TCP = auto()
    UDP = auto()
    ESP = auto()
    GRE = auto()
    SCTP = auto()


class MergePolicy(Enum):
    INVALID = 0
    OVERRIDE_EXISTING = auto()
    KEEP_ORIGINAL = auto()
    SET_UNION = auto()
    SET_UNION_WITH_MAX = auto()


@dataclass
class Flow:
    src_ip: str
    dst_ip: str
    src_port: int
    dst_port: int
    protocol: Protocol


@dataclass
class Score:
    confidence: int
    severity: int


@dataclass
class UnionValue:
    pass


@dataclass
class ListNumber(UnionValue):
    list_number_value: list[float]


@dataclass
class ListString(UnionValue):
    list_string_value: list[str]


@dataclass
class ListBoolean(UnionValue):
    list_bool_value: list[bool]


@dataclass
class JsonValue(UnionValue):
    json_value: str


@dataclass
class ListFlow(UnionValue):
    list_flow_value: list[Flow]


@dataclass
class EventMetadata:
    kv_store: dict[str, UnionValue]


@dataclass
class MergePlan:
    dedup_key: str
    metadata_key_merge_operation: dict[str, MergePolicy]


@dataclass
class Compute:
    compute_type: ComputeType
    ip_address: str | None
    compute_id: str
    fqdn: str | None
    site_id: str
    ip_mask: IpMask


@dataclass
class DataFlow:
    src_compute: Compute
    dst_compute: Compute
    transport_protocols: list[Protocol] | None
    application_protocols: list[str] | None
    server_ports: list[int] | None
    # milliseconds till epoch
    start_time: int
    # milliseconds till epoch
    end_time: int


@dataclass
class EventInfo:
    version: str
    detector_name: DetectorType
    detection_start_time_epoch_milli: int
    detection_end_time_epoch_milli: int
    score: Score
    event_metadata: EventMetadata
    merge_plan: MergePlan
    mitre_tactic: Tactic
    mitre_technique: Technique
    silent: bool
    site_ids: list[str]
    org_id: str
    affected_compute: Compute | None
    is_attacker: bool | None
    is_client: bool | None
    data_flows: list[DataFlow] | None


class Data:

    logger = logging.get_common_logger(__name__)

    def __init__(self, kafka_producer: Producer):
        self._producer = kafka_producer

    def _delivery_report(self, err: KafkaError, msg: Message) -> None:
        if err is not None:
            Data.logger.warn(f"delivery failed for event: {err}")
        else:
            Data.logger.info(
                f"event successfully produced to {msg.topic()} [{msg.partition()}] with offset {msg.offset()}"
            )

    def produce_events(self, events: list[EventInfo]) -> None:
        for event in [self._info_to_proto(event) for event in events]:
            self._producer.poll(0.0)
            try:
                self._producer.produce(
                    topic=Config.KAFKA_TOPIC, value=event.SerializeToString(), on_delivery=self._delivery_report
                )
            except ValueError:
                Data.logger.warn(f"invalid event {event}, this record will not be written to the topic")
                continue

        self._producer.flush()

    def _info_to_proto(self, event: EventInfo) -> nta_pb2.Event:
        score = nta_pb2.Score(confidence=event.score.confidence, severity=event.score.severity)

        event_metadata = nta_pb2.EventMetadata(
            kv_store={
                key: self._union_value_to_proto(cast(UnionValue, event.event_metadata.kv_store[key]))
                for key in event.event_metadata.kv_store
            }
        )

        merge_plan = nta_pb2.MergePlan(
            dedup_key=event.merge_plan.dedup_key,
            metadata_key_merge_operation={
                key: nta_pb2.MergePolicy.Name(event.merge_plan.metadata_key_merge_operation[key].value)
                for key in event.merge_plan.metadata_key_merge_operation
            },  # type: ignore
        )

        affected_compute = None
        if event.affected_compute is not None:
            affected_compute = Data._compute_to_proto(event.affected_compute)

        proto_event = nta_pb2.Event(
            version=event.version,
            detector_name=event.detector_name.name,
            detection_start_time_epoch_milli=event.detection_start_time_epoch_milli,
            detection_end_time_epoch_milli=event.detection_end_time_epoch_milli,
            score=score,
            event_metadata=event_metadata,
            merge_plan=merge_plan,
            mitre_tactic=event.mitre_tactic.name,
            mitre_technique=event.mitre_technique.name,
            silent=event.silent,
            site_ids=event.site_ids,
            org_id=event.org_id,
            affected_compute=affected_compute,
        )
        if event.is_attacker is not None:
            proto_event.is_attacker = event.is_attacker
        if event.is_client is not None:
            proto_event.is_client = event.is_client
        if event.data_flows is not None:
            for data_flow in event.data_flows[: Config.MAX_DATA_FLOWS]:
                data_flow_msg = Data._data_flow_to_proto(data_flow)
                if data_flow_msg:
                    proto_event.data_flows.append(data_flow_msg)

        return proto_event

    @staticmethod
    def _union_value_to_proto(union_value: UnionValue) -> nta_core_types_pb2.UnionValue:
        match union_value.__class__.__name__:
            case ListNumber.__name__:
                return nta_core_types_pb2.UnionValue(
                    list_number_value=nta_core_types_pb2.ListNumber(
                        list_number_value=cast(ListNumber, union_value).list_number_value
                    )
                )
            case ListString.__name__:
                return nta_core_types_pb2.UnionValue(
                    list_string_value=nta_core_types_pb2.ListString(
                        list_string_value=cast(ListString, union_value).list_string_value
                    )
                )
            case ListBoolean.__name__:
                return nta_core_types_pb2.UnionValue(
                    list_bool_value=nta_core_types_pb2.ListBoolean(
                        list_boolean_value=cast(ListBoolean, union_value).list_bool_value
                    )
                )
            case JsonValue.__name__:
                return nta_core_types_pb2.UnionValue(json_value=cast(JsonValue, union_value).json_value)
            case ListFlow.__name__:
                return nta_core_types_pb2.UnionValue(
                    list_flow_value=nta_core_types_pb2.ListFlow(
                        list_flow_value=[
                            nta_core_types_pb2.Flow(
                                src_ip=val.src_ip,
                                dst_ip=val.dst_ip,
                                src_port=val.src_port,
                                dst_port=val.dst_port,
                                protocol=core_types_pb2.ProtocolType.Name(val.protocol.value),
                            )
                            for val in cast(ListFlow, union_value).list_flow_value
                        ]
                    )
                )
            case _:
                raise UnionValueTypeError(f"unsupported UnionValueType {type(union_value)}")

    @staticmethod
    def _compute_to_proto(compute: Compute) -> core_types_pb2.ComputeMsg | None:
        compute_id = Data._uuid_str_to_message(compute.compute_id)
        ip_address = Data._ip_address_str_to_message(compute.ip_address)
        if (
            compute.compute_type == ComputeType.COMPUTE_TYPE_VM
            and not compute_id
            or compute.compute_type == ComputeType.COMPUTE_TYPE_IP
            and not ip_address
        ):
            # VM computes must have a valid UUID and external IPs must have a valid IP
            return None

        compute_message = core_types_pb2.ComputeMsg(
            compute_type=core_types_pb2.ComputeType.Name(compute.compute_type.value),
            ip_address=ip_address,
            compute_id=compute_id,
            site_id=Data._uuid_str_to_message(compute.site_id),
            ip_mask=core_types_pb2.IpMaskType.Name(compute.ip_mask.value),
        )

        if compute.fqdn is not None:
            compute_message.fqdn = core_types_pb2.FqdnMsg(compute.fqdn)

        return compute_message

    @staticmethod
    def _data_flow_to_proto(data_flow: DataFlow) -> nta_pb2.DataFlow | None:
        src_compute = Data._compute_to_proto(data_flow.src_compute)
        dst_compute = Data._compute_to_proto(data_flow.dst_compute)

        if not src_compute or not dst_compute:
            # Data flows must have valid src and dst computes
            return None

        data_flow_message = nta_pb2.DataFlow(
            src_compute=Data._compute_to_proto(data_flow.src_compute),
            dst_compute=Data._compute_to_proto(data_flow.dst_compute),
            start_time=data_flow.start_time,
            end_time=data_flow.end_time,
        )

        if data_flow.transport_protocols is not None:
            data_flow_message.transport_protocols.extend(
                [
                    core_types_pb2.ProtocolType.Name(protocol.value)
                    for protocol in data_flow.transport_protocols
                ]
            )

        if data_flow.application_protocols is not None:
            data_flow_message.application_protocols.extend(data_flow.application_protocols)

        if data_flow.server_ports is not None:
            data_flow_message.server_ports.extend(data_flow.server_ports)

        return data_flow_message

    @staticmethod
    def _ip_address_str_to_message(address: str | None) -> core_types_pb2.IpAddressMsg | None:
        if address is None:
            return None

        parsed_address = None
        try:
            parsed_address = ip_address(address)
        except ValueError as e:
            Data.logger.warn(f"failed to parse IP address with value {address}, got error {e}")
            return None

        if parsed_address is None:
            Data.logger.warn(f"failed to parse IP address with value {address}, got None object")
            return None

        if parsed_address.version == 4:
            return core_types_pb2.IpAddressMsg(
                address_family=core_types_pb2.ADDRESS_FAMILY_TYPE_V4, ipv4=int(parsed_address)
            )
        if parsed_address.version == 6:
            return core_types_pb2.IpAddressMsg(
                address_family=core_types_pb2.ADDRESS_FAMILY_TYPE_V6, ipv6=parsed_address.packed
            )

    @staticmethod
    def _uuid_str_to_message(uuid: str) -> core_types_pb2.UuidMsg | None:
        parsed_uuid = None
        try:
            parsed_uuid = UUID(uuid)
        except ValueError as e:
            Data.logger.warn(f"failed to parse UUID with value {uuid}, got error {e}")
            return None

        if parsed_uuid is None:
            Data.logger.warn(f"failed to parse UUID with value {uuid}, got None object")
            return None

        return core_types_pb2.UuidMsg(
            left=int.from_bytes(parsed_uuid.bytes[:8]), right=int.from_bytes(parsed_uuid.bytes[8:])
        )
