from collections.abc import Callable
from typing import TypeVar

from common.utils import logging
from common.utils.flow import data
from common.utils.flow.config import Config
from pandas import DataFrame
from pydruid.client import PyDruid

T = TypeVar("T")


class FlowQueryError(Exception):
    pass


class FlowService:
    logger = logging.get_common_logger(__name__)

    def __init__(self):
        self._data = data.Data(PyDruid(Config.DRUID_HOST, "druid/v2", cafile=Config.CA_PATH))

    def group_by(self, query: data.FlowGroupByQuery, extractor: Callable[[DataFrame], T] = lambda df: df) -> T:
        try:
            if query.limit_spec is None or query.limit_spec == dict():
                FlowService.logger.debug(
                    "no limit spec in the group by, applying default limit spec of {} rows".format(
                        Config.DEFAULT_GROUP_BY_LIMIT
                    )
                )
                query.limit_spec = {
                    "type": "default",
                    "limit": int(Config.DEFAULT_GROUP_BY_LIMIT),
                }

            return extractor(self._data.druid_group_by(query))
        except Exception as e:
            raise FlowQueryError("an error occurred while performing groupBy query for flows") from e

    def get_present_src_id_virtual_column(self) -> data.FlowVirtualColumn:
        return data.FlowVirtualColumn(
            "presentSrcVmId",
            "if ( like({0}, 'External') || like({0}, 'Internal'), {1}, {2})".format(
                "srcVmId", "originalSrcVmId", "srcVmId"
            ),
            data.ColumnOutputType.STRING,
        )

    def get_present_dst_id_virtual_column(self) -> data.FlowVirtualColumn:
        return data.FlowVirtualColumn(
            "presentDstVmId",
            "if ( like({0}, 'External') || like({0}, 'Internal'), {1}, {2})".format(
                "dstVmId", "originalDstVmId", "dstVmId"
            ),
            data.ColumnOutputType.STRING,
        )
