import dataclasses
import json
import urllib.error  # needed for pydruid  # noqa: F401
import urllib.request  # needed for pydruid  # noqa: F401
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, auto
from typing import Any

from common.utils import logging
from common.utils.common_model import Interval
from pandas import DataFrame
from pydruid.client import PyDruid
from pydruid.query import QueryBuilder
from pydruid.utils.dimensions import DimensionSpec
from pydruid.utils.filters import Filter
from pydruid.utils.having import Having
from pydruid.utils.postaggregator import Postaggregator


class DruidQueryEncoder(json.JSONEncoder):
    """
    Special encoder for handling datatime conversion (via isoformat())
    """

    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)


class ColumnOutputType(Enum):
    LONG = auto()
    FLOAT = auto()
    DOUBLE = auto()
    STRING = auto()
    ARRAY = auto()
    COMPLEX = auto()


@dataclass
class FlowVirtualColumn:
    name: str
    expression: str
    output_type: ColumnOutputType = ColumnOutputType.LONG
    type: str = "expression"

    def to_dict(self) -> dict[str, Any]:
        d = dataclasses.asdict(self)
        d["output_type"] = self.output_type.name
        return d


@dataclass
class FlowGroupByQuery:
    datasource: str
    granularity: str
    intervals: list[Interval]
    aggregations: dict[str, dict]
    dimensions: list[str] | list[DimensionSpec]
    filter: Filter | None = None
    having: Having | None = None
    post_aggregations: dict[str, Postaggregator] = field(default_factory=dict)
    context: dict = field(default_factory=dict)
    limit_spec: dict = field(default_factory=dict)
    virtual_columns: list[FlowVirtualColumn] = field(default_factory=list)

    def pprint(self) -> str:
        # simple pretty print of query, the query is assumed to be groupBy
        builder = QueryBuilder()
        flow_query = builder.build_query("groupBy", dataclasses.asdict(self))

        # dumps need a special json encoder to handle DateTime
        # print(json.dumps(flow_query.query_dict, cls=DruidQueryEncoder).encode("utf-8"))
        # print(flow_query.query_dict)
        return str(flow_query.query_dict)


class Data:
    logger = logging.get_common_logger(__name__)

    def __init__(self, druid_client: PyDruid):
        self._query = druid_client

    def druid_group_by(self, query: FlowGroupByQuery) -> DataFrame:
        Data.logger.debug(f"executing groupBy query on druid, query: {query}")
        self._query.groupby(
            datasource=query.datasource,
            granularity=query.granularity,
            intervals=[interval.isoformat() for interval in query.intervals],
            aggregations=query.aggregations,
            dimensions=query.dimensions,
            filter=query.filter,
            having=query.having,
            post_aggregations=query.post_aggregations,
            context=query.context,
            limit_spec=query.limit_spec,
            virtualColumns=[virtual_column.to_dict()
                            for virtual_column in query.virtual_columns]
        )
        result = self._query.export_pandas()
        Data.logger.debug(f"groupBy query retrieved {len(result.index)} rows")

        return result
