import ipaddress
import uuid
from datetime import date, datetime

import pandas as pd
from pydruid.utils.aggregators import stringlast
from pydruid.utils.filters import Dimension, Filter

from common.utils import logging
from common.utils.common_model import Interval
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery
from common.utils.lookup.config import Config
from common.utils.lookup.data import Data


IP_PREFIX = "IP-"


class LookupError(Exception):
    pass


class LookupService:
    logger = logging.get_common_logger(__name__)

    def __init__(self, flow_service: FlowService):
        self._data = Data()
        self._flow_service = flow_service

    @staticmethod
    def is_valid_ip_address(address: str) -> bool:
        try:
            ipaddress.ip_address(LookupService._check_and_remove_ip_prefix(address))
        except ValueError:
            return False
        return True

    @staticmethod
    def _check_and_remove_ip_prefix(address: str) -> str:
        ip_prefix_len = len(IP_PREFIX)
        if len(address) < ip_prefix_len:
            # address is too short to have an ip prefix
            return address
        if address[:ip_prefix_len] == IP_PREFIX:
            return address[ip_prefix_len:]
        return address

    def groups_to_computes_lookup(self, site_id: str, group_ids: list[str]) -> set[str]:
        """Find the members of the given groups"""
        now = datetime.now()
        return {
            config.compute_id
            for group_id in group_ids
            for config in self._data.find_associated_computes_by_interval_and_site(
                group_id, Interval(now, now), site_id
            )
        }

    def computes_to_ip_lookup(self, site_id: str, interval: Interval, compute_ids: list[str] = []) -> dict[str, str]:
        """Perform compute to IP lookup, if the compute_ids list is empty, lookup ips for all computes in the given site"""
        filter = Dimension("srcSiteId") == site_id
        if compute_ids != []:
            compute_ids_filter = Filter(type="in", dimension="srcVmId", values=compute_ids)
            filter = Filter(type="and", fields=[filter, compute_ids_filter])

        src_query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[interval],
            aggregations={"ipAddress": stringlast("srcIps")},
            dimensions=["srcVmId"],
            context={"queryId": f"NTA:IpLookup:{uuid.uuid4()}"},
            filter=filter,
        )

        src_result = self._flow_service.group_by(src_query).dropna()
        if compute_ids != [] and len(src_result.index) == len(compute_ids):
            if len(src_result.index) == 0:
                return {}
            return pd.Series(src_result["ipAddress"].values, index=src_result["srcVmId"]).to_dict()

        filter = Dimension("dstSiteId") == site_id
        if compute_ids != []:
            compute_ids_filter = Filter(type="in", dimension="dstVmId", values=compute_ids)
            filter = Filter(type="and", fields=[filter, compute_ids_filter])

        dst_query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[Interval(datetime.combine(date.today(), datetime.min.time()), datetime.now())],
            aggregations={"ipAddress": stringlast("dstIps")},
            dimensions=["dstVmId"],
            context={"queryId": f"NTA:IpLookup:{uuid.uuid4()}"},
            filter=filter,
        )

        dst_result = self._flow_service.group_by(dst_query).dropna()

        src_result = src_result.rename(columns={"srcVmId": "vmId"})
        dst_result = dst_result.rename(columns={"dstVmId": "vmId"})
        combined_result = pd.concat([src_result, dst_result], ignore_index=True).drop_duplicates("vmId", keep="last")

        if len(combined_result.index) == 0:
            return {}
        return pd.Series(combined_result["ipAddress"].values, index=combined_result["vmId"]).to_dict()
