from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
import json
from typing import Any

import psycopg
from common.utils import logging
from common.utils.userconfig.config import Config
from psycopg import sql
from psycopg.rows import class_row


class UserConfigConversionError(Exception):
    pass


class SettingName(Enum):
    SENSITIVITY = 0
    FIREWALL_RULE = auto()
    AGGREGATION_INTERVAL = auto()


class Setting(ABC):
    @abstractmethod
    def get_setting_name(self) -> SettingName:
        pass


@dataclass
class SensitivityMeterSetting(Setting):
    sensitivity: float

    def get_setting_name(self) -> SettingName:
        return SettingName.SENSITIVITY


@dataclass
class FirewallRuleSetting(Setting):
    check_firewall_rule: bool

    def get_setting_name(self) -> SettingName:
        return SettingName.FIREWALL_RULE


class AggregationIntervalType(Enum):
    RELATIVE_HOUR = 0
    UTC_DAY = auto()
    UTC_HOUR = auto()


@dataclass
class EventStoreAggregationIntervalSetting(Setting):
    aggregationIntervalType: AggregationIntervalType

    def get_setting_name(self) -> SettingName:
        return SettingName.AGGREGATION_INTERVAL


@dataclass
class DetectorSettings:
    settings: list[Setting]

    def get_firewall_rule_setting(self) -> FirewallRuleSetting | None:
        for setting in self.settings:
            if setting.get_setting_name() == SettingName.FIREWALL_RULE:
                return setting  # type: ignore

    def get_sensitivity_meter_setting(self) -> SensitivityMeterSetting | None:
        for setting in self.settings:
            if setting.get_setting_name() == SettingName.SENSITIVITY:
                return setting  # type: ignore

    def get_event_store_aggregation_interval_setting(self) -> EventStoreAggregationIntervalSetting | None:
        for setting in self.settings:
            if setting.get_setting_name() == SettingName.AGGREGATION_INTERVAL:
                return setting  # type: ignore


class ExclusionLabel(Enum):
    NONE = 0
    SOURCE = auto()
    DESTINATION = auto()
    RESPONDER = auto()


class ExclusionTypeInfo(Enum):
    VM = 0
    GROUP = auto()


@dataclass
class ExcludedItemsInfo:
    value: Any


@dataclass
class AnomalyExclusionSetting:
    label: ExclusionLabel
    excluded_items_info: ExcludedItemsInfo
    type: ExclusionTypeInfo


@dataclass
class AnomalyExclusionSettings:
    exclusions: list[AnomalyExclusionSetting]


class DetectorStatusLabel(Enum):
    NOT_STARTED = 0
    SUCCESS = auto()
    NOT_ENOUGH_BASELINE = auto()
    NOT_SCHEDULED = auto()
    FAILURE = auto()


@dataclass
class DetectorRunStatus:
    status_label: DetectorStatusLabel
    last_successful_run_time: int | None = None
    max_baseline_duration_days: int | None = None
    remaining_baseline_duration_days: int | None = None
    checkpoint_time: int | None = None
    message: str = ""

    def _to_update_dict(self) -> dict[str, str | int | bool]:
        result: dict[str, str | int | bool] = {}

        if self.status_label is not None:
            result["statusLabel"] = self.status_label.name
        if self.last_successful_run_time is not None:
            result["lastSuccessfulRuntime"] = self.last_successful_run_time
        if self.max_baseline_duration_days is not None:
            result["maxBaselineDurationDays"] = self.max_baseline_duration_days
        if self.remaining_baseline_duration_days is not None:
            result["remainingBaselineDurationDays"] = self.remaining_baseline_duration_days
        if self.checkpoint_time is not None:
            result["checkpointTime"] = self.checkpoint_time
        if self.message is not None:
            result["message"] = self.message

        return result


# ID is a composite of site_id, name, and org_id
@dataclass
class UserConfigEntity:
    site_id: str
    name: str
    org_id: str
    version: str
    last_modified_time: int
    enabled: bool
    silent: bool
    source: str
    block: bool
    blockconfidencethreshold: int
    mitre_tactics: list[str]
    mitre_techniques: list[str]
    settings: dict[str, Any]
    exclusion_settings: dict[str, Any]
    run_status: dict[str, Any]


@dataclass
class UserConfigInfo:
    site_id: str
    name: str
    org_id: str
    version: str
    last_modified_time: int
    enabled: bool
    silent: bool
    source: str
    block: bool
    block_confidence_threshold: int
    mitre_tactics: list[str]
    mitre_techniques: list[str]
    settings: DetectorSettings
    exclusion_settings: AnomalyExclusionSettings
    run_status: DetectorRunStatus


class Data:
    logger = logging.get_common_logger(__name__)

    def get_user_config(self, where: dict[str, str] = {}) -> list[UserConfigInfo]:
        connection_string = "dbname={dbname} user={user} password={password} host={host} port={port}".format(
            dbname=Config.POSTGRES_DB_NAME,
            user=Config.POSTGRES_USER,
            password=Config.POSTGRES_PASSWORD,
            host=Config.POSTGRES_HOST,
            port=Config.POSTGRES_PORT,
        )
        with psycopg.connect(connection_string) as conn:
            with conn.cursor(row_factory=class_row(UserConfigEntity)) as cur:
                query_string = "select * from {}"
                for i in range(len(where)):
                    if i == 0:
                        query_string += " where {} = %s"
                    else:
                        query_string += " and {} = %s"
                query_fill = [sql.Identifier(Config.CONFIG_TABLE_NAME)] + [sql.Identifier(key) for key in where.keys()]
                query = sql.SQL(query_string).format(*query_fill)

                cur.execute(query, list(where.values()))
                user_config_infos = [self._entity_to_info(record) for record in cur]
                Data.logger.info(
                    f"retrieved user config records for sites: {[info.site_id for info in user_config_infos]}"
                )
                return user_config_infos

    def update_detector_status(self, status: DetectorRunStatus, where: dict[str, str] = {}) -> None:
        connection_string = "dbname={dbname} user={user} password={password} host={host} port={port}".format(
            dbname=Config.POSTGRES_DB_NAME,
            user=Config.POSTGRES_USER,
            password=Config.POSTGRES_PASSWORD,
            host=Config.POSTGRES_HOST,
            port=Config.POSTGRES_PORT,
        )
        json_status = status._to_update_dict()
        with psycopg.connect(connection_string) as conn:
            with conn.cursor(row_factory=class_row(UserConfigEntity)) as cur:
                query_string = "update {} set run_status = run_status || {}"
                for i in range(len(where)):
                    if i == 0:
                        query_string += " where {} = %s"
                    else:
                        query_string += " and {} = %s"

                query_fill = [sql.Identifier(Config.CONFIG_TABLE_NAME), json.dumps(json_status)] + [
                    sql.Identifier(key) for key in where.keys()
                ]
                query = sql.SQL(query_string).format(*query_fill)
                Data.logger.debug(f"query string : {query}")
                cur.execute(query, list(where.values()))

    def _entity_to_info(self, entity: UserConfigEntity) -> UserConfigInfo:
        return UserConfigInfo(
            entity.site_id,
            entity.name,
            entity.org_id,
            entity.version,
            entity.last_modified_time,
            entity.enabled,
            entity.silent,
            entity.source,
            entity.block,
            entity.blockconfidencethreshold,
            entity.mitre_tactics,
            entity.mitre_techniques,
            self._convert_settings_dict(entity.settings),
            self._convert_exclusion_settings_dict(entity.exclusion_settings),
            self._convert_run_status_dict(entity.run_status),
        )

    def _convert_settings_dict(self, settings: dict[str, Any]) -> DetectorSettings:
        if "detectorSettings" not in settings:
            raise UserConfigConversionError("invalid detector settings, does not contain detectorSettings key")

        converted_settings = DetectorSettings([])
        for setting in settings["detectorSettings"]:
            if setting["discriminator"] == "SENSITIVITY":
                converted_settings.settings.append(SensitivityMeterSetting(setting["sensitivity"]))
            elif setting["discriminator"] == "FIREWALL_RULE":
                converted_settings.settings.append(FirewallRuleSetting(setting["checkFirewallRule"]))
            elif setting["discriminator"] == "AGGREGATION_INTERVAL":
                converted_settings.settings.append(
                    EventStoreAggregationIntervalSetting(AggregationIntervalType[setting["aggregationIntervalType"]])
                )
        return converted_settings

    def _convert_exclusion_settings_dict(self, settings: dict[str, Any]) -> AnomalyExclusionSettings:
        if "exclusions" not in settings:
            raise UserConfigConversionError("invalid exclusion settings, does not contain exclusions key")

        converted_settings = AnomalyExclusionSettings([])
        for exclusion in settings["exclusions"]:
            exclusion = AnomalyExclusionSetting(
                label=ExclusionLabel[exclusion["label"]],
                excluded_items_info=ExcludedItemsInfo(value=exclusion["excludedItemsInfo"]["excludedItems"]),
                type=ExclusionTypeInfo[exclusion["type"]],
            )
            converted_settings.exclusions.append(exclusion)

        return converted_settings

    def _convert_run_status_dict(self, run_status: dict[str, Any]) -> DetectorRunStatus:
        return DetectorRunStatus(
            status_label=getattr(DetectorStatusLabel, run_status["statusLabel"]),
            last_successful_run_time=run_status["lastSuccessfulRuntime"],
            max_baseline_duration_days=run_status["maxBaselineDurationDays"],
            remaining_baseline_duration_days=run_status["remainingBaselineDurationDays"],
            checkpoint_time=run_status["checkpointTime"],
            message=run_status["message"],
        )
